<?php

namespace App\Http\Controllers;

use App\MetaGer;
use App\PrometheusExporter;
use Illuminate\Support\Facades\Cache;
use Illuminate\Http\Request;
use Mcamara\LaravelLocalization\Facades\LaravelLocalization;
use Illuminate\Support\Facades\Log;
use Prometheus\CollectorRegistry;
use Illuminate\Support\Facades\View;
use Illuminate\Support\Facades\App;
use App\QueryTimer;
use App\SearchSettings;

class MetaGerSearch extends Controller
{
    public function search(Request $request, MetaGer $metager, $timing = false)
    {
        $query_timer = \app()->make(QueryTimer::class);
        $locale = LaravelLocalization::getCurrentLocale();
        $preferredLanguage = array($request->getPreferredLanguage());
        if (!empty($preferredLanguage) && !empty($locale)) {
            PrometheusExporter::PreferredLanguage($locale, $preferredLanguage);
        }

        if ($request->filled("chrome-plugin")) {
            return redirect(LaravelLocalization::getLocalizedURL(LaravelLocalization::getCurrentLocale(), "/plugin"));
        }

        $focus = $request->input("focus", "web");

        if ($focus === "maps") {
            $searchinput = $request->input('eingabe', '');
            return redirect()->to('https://maps.metager.de/map/' . $searchinput . '/1240908.5493525574,6638783.2192695495,6');
        }

        # If there is no query parameter we redirect to the startpage
        $eingabe = $request->input('eingabe', '');
        if (empty(trim($eingabe))) {
            return redirect(LaravelLocalization::getLocalizedURL(LaravelLocalization::getCurrentLocale(), '/'));
        }

        # Mit gelieferte Formulardaten parsen und abspeichern:
        $query_timer->observeStart("Search_ParseFormData");
        $metager->parseFormData($request);
        $query_timer->observeEnd("Search_ParseFormData");

        # Nach Spezialsuchen überprüfen:
        $query_timer->observeStart("Search_CheckSpecialSearches");
        $metager->checkSpecialSearches($request);
        $query_timer->observeEnd("Search_CheckSpecialSearches");

        # Search query can be empty after parsing the formdata
        # we will cancel the search in that case and show an error to the user
        if (empty($metager->getQ())) {
            return $metager->createView();
        }

        $query_timer->observeStart("Search_CreateQuicktips");
        $quicktips = $metager->createQuicktips();
        $query_timer->observeEnd("Search_CreateQuicktips");

        # Suche für alle zu verwendenden Suchmaschinen als Job erstellen,
        # auf Ergebnisse warten und die Ergebnisse laden
        $query_timer->observeStart("Search_CreateSearchEngines");
        $metager->createSearchEngines($request);
        $query_timer->observeEnd("Search_CreateSearchEngines");

        $query_timer->observeStart("Search_StartSearch");
        $metager->startSearch();
        $query_timer->observeEnd("Search_StartSearch");

        # Versuchen die Ergebnisse der Quicktips zu laden
        if ($quicktips !== null) {
            $query_timer->observeStart("Search_LoadQuicktips");
            $quicktipResults = $quicktips->loadResults();
            $query_timer->observeEnd("Search_LoadQuicktips");
        } else {
            $quicktipResults = [];
        }

        $query_timer->observeStart("Search_WaitForMainResults");
        $metager->waitForMainResults();
        $query_timer->observeEnd("Search_WaitForMainResults");

        $query_timer->observeStart("Search_RetrieveResults");
        $metager->retrieveResults();
        $query_timer->observeEnd("Search_RetrieveResults");

        # Alle Ergebnisse vor der Zusammenführung ranken:
        $query_timer->observeStart("Search_RankAll");
        $metager->rankAll();
        $query_timer->observeEnd("Search_RankAll");

        # Ergebnisse der Suchmaschinen kombinieren:
        $query_timer->observeStart("Search_PrepareResults");
        $metager->prepareResults();
        $query_timer->observeEnd("Search_PrepareResults");
        //        $admitad = [];
        $adgoal = [];

        $query_timer->observeStart("Search_Affiliates");
        if (!$metager->isApiAuthorized() && !$metager->isDummy()) {
            /*
            $newAdmitad = new \App\Models\Admitad($metager);
            if (!empty($newAdmitad->hash)) {
                $admitad[] = $newAdmitad;
            }*/
            $newAdgoal = new \App\Models\Adgoal($metager);
            if (!empty($newAdgoal->hash)) {
                $adgoal[] = $newAdgoal;
            }
        }

        // $metager->parseAffiliates($admitad);
        $metager->parseAffiliates($adgoal);

        // Add Advertisement for Donations
        if (!$metager->isApiAuthorized() && !$metager->isDummy()) {
            $metager->addDonationAdvertisement();
        }
        $query_timer->observeEnd("Search_Affiliates");

        foreach ($metager->getEngines() as $engine) {
            if ($engine->loaded) {
                $engine->setNew(false);
                $engine->markNew();
            }
        }
        $query_timer->observeStart("Search_CacheFiller");
        try {
            Cache::put("loader_" . $metager->getSearchUid(), [
                "metager" => [
                    "apiAuthorized" => $metager->isApiAuthorized(),
                ],
                //                "admitad" => $admitad,
                "adgoal" => $adgoal,
                "engines" => $metager->getEngines(),
            ], 60 * 60);
        } catch (\Exception $e) {
            Log::error($e->getMessage());
        }
        $query_timer->observeEnd("Search_CacheFiller");

        # Die Ausgabe erstellen:
        $resultpage = $metager->createView($quicktipResults);

        $registry = CollectorRegistry::getDefault();
        $counter = $registry->getOrRegisterCounter('metager', 'result_counter', 'counts total number of returned results', []);
        $counter->incBy(sizeof($metager->getResults()));
        $counter = $registry->getOrRegisterCounter('metager', 'query_counter', 'counts total number of search queries', []);
        $counter->inc();

        $query_timer->observeTotal();
        // Splitting the response return into multiple parts.
        // This might speed up page view time for users with slow network
        $responseArray = str_split($resultpage->render(), 1024);
        foreach ($responseArray as $responsePart) {
            echo ($responsePart);
            flush();
        }
    }

    public function searchTimings(Request $request, MetaGer $metager)
    {
        $request->merge([
            'eingabe' => "Hannover",
        ]);
        return $this->search($request, $metager, true);
    }

    public function loadMore(Request $request)
    {
        /**
         * There are three forms of requests to the resultpage
         * 1. Initial Request: Loads the fastest searchengines and sends them to the user
         * 2. Load more results (with JS): Loads new search engines that answered after the initial request was send
         * 3. Load more results (without JS): Loads new search engines that answered within 1s timeout
         */
        if ($request->filled('loadMore') && $request->filled('script') && $request->input('script') === "yes") {
            return $this->loadMoreJS($request);
        }
    }

    private function loadMoreJS(Request $request)
    {
        \app()->make(SearchSettings::class)->javascript_enabled = true;
        # Create a MetaGer Instance with the supplied hash
        $hash = $request->input('loadMore', '');

        # Parser Skripte einhängen
        $dir = app_path() . "/Models/parserSkripte/";
        foreach (scandir($dir) as $filename) {
            $path = $dir . $filename;
            if (is_file($path)) {
                require_once $path;
            }
        }

        $cached = Cache::get($hash);
        if ($cached === null) {
            return response()->json(['finished' => true]);
        }

        $engines = $cached["engines"];
        $adgoal = $cached["adgoal"];
        //        $admitad = $cached["admitad"];
        $mg = $cached["metager"];

        $metager = new MetaGer(substr($hash, strpos($hash, "loader_") + 7));
        $metager->setApiAuthorized($mg["apiAuthorized"]);

        $metager->parseFormData($request, false);
        # Nach Spezialsuchen überprüfen:
        $metager->checkSpecialSearches($request);
        $metager->restoreEngines($engines);

        # Checks Cache for engine Results
        $metager->checkCache();

        $metager->retrieveResults();

        $metager->rankAll();
        $metager->prepareResults();

        if (!$metager->isApiAuthorized() && !$metager->isDummy()) {
            /*$newAdmitad = new \App\Models\Admitad($metager);
            if (!empty($newAdmitad->hash)) {
                $admitad[] = $newAdmitad;
            }*/
            $newAdgoal = new \App\Models\Adgoal($metager);
            if (!empty($newAdgoal->hash)) {
                $adgoal[] = $newAdgoal;
            }
        }

        //        $admitadFinished = $metager->parseAffiliates($admitad);
        $adgoalFinished = $metager->parseAffiliates($adgoal);

        $result = [
            'finished' => true,
            'newResults' => [],
            'changedResults' => [],
        ];
        $result["nextSearchLink"] = $metager->nextSearchLink();

        $newResults = 0;
        foreach ($metager->getResults() as $index => $resultTmp) {
            if ($resultTmp->new || $resultTmp->changed) {
                if ($metager->getFokus() !== "bilder") {
                    $view = View::make('layouts.result', ['index' => $index, 'result' => $resultTmp, 'metager' => $metager]);
                    $html = $view->render();
                    if (!$resultTmp->new && $resultTmp->changed) {
                        $result['changedResults'][$index] = $html;
                    } else {
                        $result['newResults'][$index] = $html;
                    }
                    $result["imagesearch"] = false;
                } else {
                    $view = View::make('layouts.image_result', ['index' => $index, 'result' => $resultTmp, 'metager' => $metager]);
                    $html = $view->render();
                    if (!$resultTmp->new && $resultTmp->changed) {
                        $result['changedResults'][$index] = $html;
                    } else {
                        $result['newResults'][$index] = $html;
                    }
                    $result["imagesearch"] = true;
                }
                $newResults++;
            }
        }

        $finished = true;
        $enginesLoaded = [];
        foreach ($engines as $engine) {
            if (!$engine->loaded) {
                $enginesLoaded[$engine->name] = false;
                $finished = false;
            } else {
                $enginesLoaded[$engine->name] = true;
                $engine->setNew(false);
                $engine->markNew();
            }
        }

        if (!$adgoalFinished /*|| !$admitadFinished*/) {
            $finished = false;
        }

        $result["finished"] = $finished;
        $result["engines"] = $enginesLoaded;

        if ($newResults > 0) {
            $registry = CollectorRegistry::getDefault();
            $counter = $registry->getOrRegisterCounter('metager', 'result_counter', 'counts total number of returned results', []);
            $counter->incBy($newResults);
        }
        // Update new Engines
        Cache::put("loader_" . $metager->getSearchUid(), [
            "metager" => [
                "apiAuthorized" => $metager->isApiAuthorized(),
            ],
            //            "admitad" => $admitad,
            "adgoal" => $adgoal,
            "engines" => $metager->getEngines(),
        ], 1 * 60);

        # JSON encoding will fail if invalid UTF-8 Characters are in this string
        # mb_convert_encoding will remove thise invalid characters for us
        $result = mb_convert_encoding($result, "UTF-8", "UTF-8");
        return response()->json($result);
    }

    public function botProtection($redirect)
    {
        $hash = md5(date('YmdHi'));
        return view('botProtection')
            ->with('hash', $hash)
            ->with('r', $redirect);
    }

    public function get($url)
    {
        $ctx = stream_context_create(array('http' => array('timeout' => 2)));
        return file_get_contents($url, false, $ctx);
    }

    private function startsWith($haystack, $needle)
    {
        $length = strlen($needle);
        return (substr($haystack, 0, $length) === $needle);
    }

    public function tips(Request $request)
    {
        $tipserver = '';
        if (App::environment() === "development") {
            $tipserver = "https://dev.quicktips.metager.de/1.1/tips.xml";
        } else {
            $tipserver = "https://quicktips.metager.de/1.1/tips.xml";
        }
        if (LaravelLocalization::getCurrentLocale() == "en") {
            $tipserver .= "?locale=en";
        }
        $tips_text = file_get_contents($tipserver);
        $tips = [];
        try {
            $tips_xml = \simplexml_load_string($tips_text);

            $tips_xml->registerXPathNamespace('mg', 'http://metager.de/tips/');
            $tips_xml = $tips_xml->xpath('mg:tip');
            foreach ($tips_xml as $tip_xml) {
                $tips[] = $tip_xml->__toString();
            }
        } catch (\Exception $e) {
            Log::error("A problem occurred loading tips from the tip server.");
            Log::error($e->getMessage());
            abort(500);
        }
        return view('tips')
            ->with('title', trans('tips.title'))
            ->with('tips', $tips);
    }
}
