<?php

namespace App\Models;

use Illuminate\Support\Facades\Redis;
use \Carbon\Carbon;

class Key
{
    public $key;
    public $status; # Null If Key invalid | false if valid but has no adFreeSearches | true if valid and has adFreeSearches
    private $keyserver = "https://key.metager.de/";
    public $keyinfo;
    const CHANGE_EVERY = 1 * 24 * 60 * 60;

    public function __construct($key, $status = null)
    {
        $this->key = $key;
        $this->status = $status;
        if (\app()->environment() !== "production") {
            $this->keyserver = "https://dev.key.metager.de/";
        }
    }

    # always returns true or false
    public function getStatus()
    {
        if ($this->key !== '' && $this->status === null) {
            $this->updateStatus();
            if ($this->status === null) {
                // The user provided an invalid key which we will log to fail2ban
                $fail2banEnabled = config("metager.metager.fail2ban.enabled");
                if (!empty($fail2banEnabled) && $fail2banEnabled && !config("metager.metager.fail2ban.url") && !config("metager.metager.fail2ban.user") && !config("metager.metager.fail2ban.password")) {
                    // Submit fetch job to worker
                    $mission = [
                        "resulthash" => "captcha",
                        "url" => config("metager.metager.fail2ban.url") . "/mgkeytry/",
                        "useragent" => "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:81.0) Gecko/20100101 Firefox/81.0",
                        "username" => config("metager.metager.fail2ban.user"),
                        "password" => config("metager.metager.fail2ban.password"),
                        "headers" => [
                            "ip" => \request()->ip()
                        ],
                        "cacheDuration" => 0,
                        "name" => "Captcha",
                    ];
                    $mission = json_encode($mission);
                    Redis::rpush(\App\MetaGer::FETCHQUEUE_KEY, $mission);
                }
            }
        }
        return $this->status;
    }

    public function updateStatus()
    {
        $authKey = base64_encode(config("metager.metager.keyserver.user") . ':' . config("metager.metager.keyserver.password"));

        $opts = array(
            'http' => array(
                'method' => 'GET',
                'header' => 'Authorization: Basic ' . $authKey,
            ),
        );
        $context = stream_context_create($opts);

        try {
            $link = $this->keyserver . "v2/key/" . urlencode($this->key);
            $result = json_decode(file_get_contents($link, false, $context));
            if (!empty($result)) {
                $this->keyinfo = $result;
                if ($this->keyinfo->adFreeSearches > 0 || $this->keyinfo->apiAccess === "unlimited") {
                    $this->status = true;
                } else {
                    $this->status = false;
                }
                return true;
            } else {
                return false;
            }
        } catch (\ErrorException $e) {
            return false;
        }
    }

    public function requestPermission()
    {
        $authKey = base64_encode(config("metager.metager.keyserver.user") . ':' . config("metager.metager.keyserver.password"));
        $postdata = http_build_query(array(
            'dummy' => 0,
        ));
        $opts = array(
            'http' => array(
                'method' => 'POST',
                'header' => [
                    'Content-type: application/x-www-form-urlencoded',
                    'Authorization: Basic ' . $authKey
                ],
                'content' => $postdata,
            ),
        );

        $context = stream_context_create($opts);

        try {
            $link = $this->keyserver . "v2/key/" . urlencode($this->key) . "/request-permission";
            $result = json_decode(file_get_contents($link, false, $context));
            if ($result->{'apiAccess'} == true) {
                return true;
            } else {
                $this->status = false;
                return false;
            }
        } catch (\ErrorException $e) {
            return false;
        }
    }
    public function generateKey($payment = null, $adFreeSearches = null, $key = null, $notes = "")
    {
        $authKey = base64_encode(config("metager.metager.keyserver.user") . ':' . config("metager.metager.keyserver.password"));
        $postdata = array(
            'apiAccess' => 'normal',
            'expiresAfterDays' => 365,
            'notes' => $notes
        );
        if (!empty($key)) {
            $postdata["key"] = $key;
        }

        if (!empty($payment)) {
            $postdata["payment"] = $payment;
        } else if (!empty($adFreeSearches)) {
            $postdata["adFreeSearches"] = $adFreeSearches;
        } else {
            return false;
        }
        $postdata = http_build_query($postdata, "", "&", PHP_QUERY_RFC3986);
        $opts = array(
            'http' => array(
                'method' => 'POST',
                'header' => [
                    'Content-type: application/x-www-form-urlencoded',
                    'Authorization: Basic ' . $authKey
                ],
                'content' => $postdata,
                'timeout' => 5
            ),
        );

        $context = stream_context_create($opts);

        try {
            $link = $this->keyserver . "v2/key/";
            $result = json_decode(file_get_contents($link, false, $context));
            return $result->{'mgKey'};
        } catch (\ErrorException $e) {
            return false;
        }
    }

    public function reduce($count)
    {
        $authKey = base64_encode(config("metager.metager.keyserver.user") . ':' . config("metager.metager.keyserver.password"));
        $postdata = http_build_query(array(
            'adFreeSearches' => $count,
        ));
        $opts = array(
            'http' => array(
                'method' => 'POST',
                'header' => [
                    'Content-type: application/x-www-form-urlencoded',
                    'Authorization: Basic ' . $authKey
                ],
                'content' => $postdata,
                'timeout' => 5
            ),
        );

        $context = stream_context_create($opts);

        try {
            $link = $this->keyserver . "v2/key/" . $this->key . "/reduce-searches";
            $result = json_decode(file_get_contents($link, false, $context));
            return $result;
        } catch (\ErrorException $e) {
            return false;
        }
    }

    /**
     * Tells if this key is liable to change to a custom key
     * Currently only members are allowed to do so and only every 2 days
     * Also only the original member key is allowed to be changed
     * 
     * @return boolean
     */
    public function canChange()
    {
        if (empty($this->status) || !preg_match("/^Mitgliederschlüssel\./", $this->keyinfo->notes) || $this->keyinfo->adFreeSearches < \App\Http\Controllers\KeyController::KEYCHANGE_ADFREE_SEARCHES) {
            return false;
        }
        if (!empty($this->keyinfo->KeyChangedAt)) {
            // "2021-03-09T09:19:44.000Z"
            $keyChangedAt = Carbon::createFromTimeString($this->keyinfo->KeyChangedAt, 'Europe/London');
            if ($keyChangedAt->diffInSeconds(Carbon::now()) > self::CHANGE_EVERY) {
                return true;
            } else {
                return false;
            }
        }
        return true;
    }

    public function checkForChange($hash, $newkey = "")
    {
        $authKey = base64_encode(config("metager.metager.keyserver.user") . ':' . config("metager.metager.keyserver.password"));
        $postdata = http_build_query(array(
            'hash' => $hash,
            'key' => $newkey,
        ));
        $opts = array(
            'http' => array(
                'method' => 'POST',
                'header' => [
                    'Content-type: application/x-www-form-urlencoded',
                    'Authorization: Basic ' . $authKey
                ],
                'content' => $postdata,
                'timeout' => 5
            ),
        );

        $context = stream_context_create($opts);

        try {
            $link = $this->keyserver . "v2/key/can-change";
            $result = json_decode(file_get_contents($link, false, $context));

            if (!empty($result) && $result->status === "success" && empty($result->results)) {
                return true;
            } else {
                return false;
            }
        } catch (\ErrorException $e) {
            return false;
        }
    }
}
